@@ -36,6 +36,27 @@ import gettext
 import os
 import re
 
+class NoneTranslations:
+    def gettext(self, message):
+        return None
+
+    def lgettext(self, message):
+        return None
+
+    def ngettext(self, msgid1, msgid2, n):
+        return None
+
+    def lngettext(self, msgid1, msgid2, n):
+        return None
+
+    def ugettext(self, message):
+        return None
+
+    def ungettext(self, msgid1, msgid2, n):
+        return None
+
+
+
 class MessageOutput:
     def __init__(self, with_translations = 0):
         self.messages = []
@@ -116,13 +137,17 @@ msgstr ""
 
 
 def normalizeNode(node):
+    #print >>sys.stderr, "<%s> (%s) [%s]" % (node.name, node.type, node.serialize('utf-8'))
     if not node:
         return
     elif isSpacePreserveNode(node):
         return
     elif node.isText():
         if node.isBlankNode():
-            node.setContent('')
+            if expand_entities or ( not (node.prev and not node.prev.isBlankNode()
+                                         and node.next and not node.next.isBlankNode()) ):
+                #print >>sys.stderr, "BLANK"
+                node.setContent('')
         else:
             node.setContent(re.sub('\s+',' ', node.content))
 
@@ -213,16 +238,40 @@ def getTranslation(text, spacepreserve = 0):
     text should be a string to look for, spacepreserve set to 1
     when spaces should be preserved.
     """
+    #print >>sys.stderr,"getTranslation('%s')" % (text.encode('utf-8'))
     text = normalizeString(text, not spacepreserve)
     if (text.strip() == ''):
         return text
     file = open(mofile, "rb")
     if file:
+        myfallback = NoneTranslations()
         gt = gettext.GNUTranslations(file)
+        gt.add_fallback(myfallback)
         if gt:
-            return gt.ugettext(text.decode('utf-8'))
+            res = gt.ugettext(text.decode('utf-8'))
+            return res
+
     return text
 
+def myAttributeSerialize(node):
+    result = ''
+    if node.children:
+        child = node.children
+        while child:
+            if child.type=='text':
+                result += doc.encodeEntitiesReentrant(child.content)
+            elif child.type=='entity_ref':
+                if not expand_entities:
+                    result += '&' + child.name + ';'
+                else:
+                    result += child.content.decode('utf-8')
+            else:
+                result += myAttributeSerialize(child)
+            child = child.next
+    else:
+        result = node.serialize('utf-8')
+    return result
+
 def startTagForNode(node):
     if not node:
         return 0
@@ -232,8 +281,11 @@ def startTagForNode(node):
     if node.properties:
         for p in node.properties:
             if p.type == 'attribute':
-                # FIXME: This part sucks
-                params += p.serialize('utf-8')
+                try:
+                    nsprop = p.ns().name + ":" + p.name
+                except:
+                    nsprop = p.name
+                params += " %s=\"%s\"" % (nsprop, myAttributeSerialize(p))
     return result+params
         
 def endTagForNode(node):
@@ -260,7 +312,7 @@ def isFinalNode(node):
         final_children = 1
         child = node.children
         while child and final_children:
-            if not isFinalNode(child):
+            if not child.isBlankNode() and child.type != 'comment' and not isFinalNode(child):
                 final_children = 0
             child = child.next
         if final_children:
@@ -302,32 +354,39 @@ def getCommentForNode(node):
     else:
         return None
 
-
 def replaceNodeContentsWithText(node,text):
     """Replaces all subnodes of a node with contents of text treated as XML."""
-    if node.children:
-        starttag = node.name #startTagForNode(node)
-        endtag = endTagForNode(node)
-        try:
-            # Lets add document DTD so entities are resolved
-            dtd = doc.intSubset()
-            tmp = ''
-            if expand_entities: # FIXME: we get a "Segmentation fault" in libxml2.parseMemory() when we include DTD otherwise
-                tmp = dtd.serialize('utf-8')
-            tmp = tmp + '<%s>%s</%s>' % (starttag, text, endtag)
-        except:
-            tmp = '<%s>%s</%s>' % (starttag, text, endtag)
 
+    if node.children:
+        starttag = startTagForNode(node)
+        endtag = endTagForNode(node)
+
+        # Lets add document DTD so entities are resolved
+        tmp = '<?xml version="1.0" encoding="utf-8" ?>'
         try:
-            ctxt = libxml2.createDocParserCtxt(tmp.encode('utf-8'))
+            dtd = doc.intSubset()
+            tmp = tmp + dtd.serialize('utf-8')
+        except libxml2.treeError:
+            pass
+            
+        content = '<%s>%s</%s>' % (starttag, text, endtag)
+        tmp = tmp + content.encode('utf-8')
+
+        newnode = None
+        try:
+            ctxt = libxml2.createDocParserCtxt(tmp)
             ctxt.replaceEntities(0)
             ctxt.parseDocument()
             newnode = ctxt.doc()
         except:
+            pass
+
+        if not newnode:
             print >> sys.stderr, """Error while parsing translation as XML:\n"%s"\n""" % (text.encode('utf-8'))
             return
 
         newelem = newnode.getRootElement()
+
         if newelem and newelem.children:
             free = node.children
             while free:
@@ -335,7 +394,11 @@ def replaceNodeContentsWithText(node,text):
                 free.unlinkNode()
                 free = next
 
-            node.addChildList(newelem.children)
+            if node:
+                copy = newelem.copyNodeList()
+                next = node.next
+                node.replaceNode(newelem.copyNodeList())
+                node.next = next
         else:
             # In practice, this happens with tags such as "<para>    </para>" (only whitespace in between)
             pass
@@ -407,17 +470,23 @@ def processElementTag(node, replacements, restart = 0):
             translation = getTranslation(outtxt, isSpacePreserveNode(node))
         else:
             translation = outtxt
+
         starttag = startTagForNode(node)
         endtag = endTagForNode(node)
 
-        if restart or worthOutputting(node):
+        worth = worthOutputting(node)
+        if not translation:
+            translation = outtxt.decode('utf-8')
+            if worth and mark_untranslated: node.setLang('C')
+
+        if restart or worth:
             i = 0
             while i < len(myrepl):
                 replacement = '<%s>%s</%s>' % (myrepl[i][0], myrepl[i][3], myrepl[i][2])
                 i += 1
                 translation = translation.replace('<placeholder-%d/>' % (i), replacement)
 
-            if worthOutputting(node):
+            if worth:
                 if mode == 'merge':
                     replaceNodeContentsWithText(node, translation)
                 else:
@@ -503,9 +572,10 @@ def read_ignoredtags(filelist):
 
 def tryToUpdate(allargs, lang):
     # Remove "-u" and "--update-translation"
+    print >>sys.stderr, "OVDI!"
     command = allargs[0]
     args = allargs[1:]
-    opts, args = getopt.getopt(args, 'avhmket:o:p:u:',
+    opts, args = getopt.getopt(args, 'avhm:ket:o:p:u:',
                                ['automatic-tags','version', 'help', 'keep-entities', 'extract-all-entities', 'merge', 'translation=',
                                 'output=', 'po-file=', 'update-translation=' ])
     for opt, arg in opts:
@@ -547,7 +617,7 @@ def tryToUpdate(allargs, lang):
             sys.stderr.write("Error: cannot rename file.\n")
             sys.exit(11)
         else:
-            os.system("msgfmt -cv -o /dev/null %s" % (file))
+            os.system("msgfmt -cv -o %s %s" % (NULL_STRING, file))
             sys.exit(0)
 
 def load_mode(modename):
@@ -581,14 +651,19 @@ mofile = ''
 ultimate = [ ]
 ignored = [ ]
 filenames = [ ]
+translationlanguage = ''
 
 mode = 'pot' # 'pot' or 'merge'
 automatic = 0
 expand_entities = 1
+mark_untranslated = 0
 expand_all_entities = 0
 
 output  = '-' # this means to stdout
 
+NULL_STRING = '/dev/null'
+if not os.path.exists('/dev/null'): NULL_STRING = 'NUL'
+
 import getopt, fileinput
 
 def usage (with_help = False):
@@ -607,6 +682,10 @@ OPTIONS may be some of:
     -r    --reuse=FILE         Specify translated XML file with the same structure
     -t    --translation=FILE   Specify MO file containing translation, and merge
     -u    --update-translation=LANG.po   Updates a PO file using msgmerge program
+
+    -l    --language=LANG      Set language of the translation to LANG
+          --mark-untranslated  Set 'xml:lang="C"' on untranslated tags
+
     -v    --version            Output version of the xml2po program
 
     -h    --help               Output this message
@@ -626,9 +705,9 @@ EXAMPLES:
 if len(sys.argv) < 2: usage()
 
 args = sys.argv[1:]
-try: opts, args = getopt.getopt(args, 'avhkem:t:o:p:u:r:',
+try: opts, args = getopt.getopt(args, 'avhkem:t:o:p:u:r:l:',
                            ['automatic-tags','version', 'help', 'keep-entities', 'expand-all-entities', 'mode=', 'translation=',
-                            'output=', 'po-file=', 'update-translation=', 'reuse=' ])
+                            'output=', 'po-file=', 'update-translation=', 'reuse=', 'language=', 'mark-untranslated' ])
 except getopt.GetoptError: usage(True)
 
 for opt, arg in opts:
@@ -638,12 +717,16 @@ for opt, arg in opts:
         automatic = 1
     elif opt in ('-k', '--keep-entities'):
         expand_entities = 0
+    elif opt in ('--mark-untranslated',):
+        mark_untranslated = 1
     elif opt in ('-e', '--expand-all-entities'):
         expand_all_entities = 1
+    elif opt in ('-l', '--language'):
+        translationlanguage = arg
     elif opt in ('-t', '--translation'):
         mofile = arg
         mode = 'merge'
-        translationlanguage = os.path.splitext(mofile)[0]
+        if translationlanguage == '': translationlanguage = os.path.split(os.path.splitext(mofile)[0])[1]
     elif opt in ('-r', '--reuse'):
         origxml = arg
     elif opt in ('-u', '--update-translation'):
@@ -651,8 +734,8 @@ for opt, arg in opts:
     elif opt in ('-p', '--po-file'):
         mofile = ".xml2po.mo"
         pofile = arg
-        translationlanguage = os.path.splitext(pofile)[0]
-        os.system("msgfmt -o %s %s >/dev/null" % (mofile, pofile)) and sys.exit(7)
+        if translationlanguage == '': translationlanguage = os.path.split(os.path.splitext(pofile)[0])[1]
+        os.system("msgfmt -o %s %s >%s" % (mofile, pofile, NULL_STRING)) and sys.exit(7)
         mode = 'merge'
     elif opt in ('-o', '--output'):
         output = arg
